# Author: Christian Brodbeck <christianbrodbeck@nyu.edu>
from inspect import getargspec
import re

from .. import testnd
from .._exceptions import DefinitionError
from .definitions import Definition
from .vardef import GroupVar


__test__ = False
TAIL_REPR = {0: '=', 1: '>', -1: '<'}


def assemble_tests(test_dict):
    "Interpret dict with test definitions"
    out = {}
    for key, params in test_dict.items():
        if isinstance(params, Test):
            out[key] = params
            continue
        elif not isinstance(params, dict):
            raise TypeError(f"Invalid object for test definition {key}: {params!r}")
        params = params.copy()
        if 'stage 1' in params:
            params['stage_1'] = params.pop('stage 1')
        kind = params.pop('kind')
        if kind in TEST_CLASSES:
            out[key] = TEST_CLASSES[kind](**params)
        else:
            raise DefinitionError(f"Unknown test kind in test definition {key}: {kind}")
    return out


def tail_arg(tail):
    try:
        if tail == 0:
            return 0
        elif tail > 0:
            return 1
        else:
            return -1
    except Exception:
        raise TypeError("tail=%r; needs to be 0, -1 or 1" % (tail,))


class Test(Definition):
    "Baseclass for any test"
    kind = None
    DICT_ATTRS = ('kind', 'model', 'vars')

    def __init__(self, desc, model, vars=None):
        self.desc = desc
        self.model = model
        self.vars = vars

        if model is None:  # no averaging
            self._between = None
            self._within_model = None
            self._within_model_items = None
        else:
            model_elements = list(map(str.strip, model.split('%')))
            if 'group' in model_elements:
                self._between = model_elements.index('group')
                del model_elements[self._between]
            else:
                self._between = None
            self._within_model_items = model_elements
            self._within_model = '%'.join(model_elements)


class EvokedTest(Test):
    "Group level test applied to subject averages"
    def __init__(self, desc, model, cat=None, vars=None):
        Test.__init__(self, desc, model, vars)
        self.cat = cat
        if cat is not None:
            if self._within_model is None or len(self._within_model_items) == 0:
                cat = None
            elif self._between is not None:
                # remove between factor from cat
                cat = [[c for i, c in enumerate(cat) if i != self._between]
                       for cat in self.cat]
        self._within_cat = cat

    def make(self, y, ds, force_permutation, kwargs):
        raise NotImplementedError


class TTestOneSample(EvokedTest):
    kind = 'ttest_1samp'
    DICT_ATTRS = Test.DICT_ATTRS + ('tail',)

    def __init__(self, tail=0):
        desc = "%s 0" % TAIL_REPR[tail]
        EvokedTest.__init__(self, desc, '')
        self.tail = tail

    def make(self, y, ds, force_permutation, kwargs):
        return testnd.ttest_1samp(
            y, match='subject', ds=ds, tail=self.tail,
            force_permutation=force_permutation, **kwargs)


class TTest(EvokedTest):
    DICT_ATTRS = Test.DICT_ATTRS + ('c1', 'c0', 'tail')

    def __init__(self, model, c1, c0, tail, vars=None):
        tail = tail_arg(tail)
        desc = '%s %s %s' % (c1, TAIL_REPR[tail], c0)
        EvokedTest.__init__(self, desc, model, (c1, c0), vars)
        self.c1 = c1
        self.c0 = c0
        self.tail = tail


class TTestInd(TTest):
    "Independent measures t-test"
    kind = 'ttest_ind'

    def __init__(self, model, c1, c0, tail=0, vars=None):
        if vars is None and model == 'group':
            vars = (('group', GroupVar((c1, c0))),)
        TTest.__init__(self, model, c1, c0, tail, vars)

    def make(self, y, ds, force_permutation, kwargs):
        return testnd.ttest_ind(
            y, self.model, self.c1, self.c0, 'subject', ds=ds, tail=self.tail,
            force_permutation=force_permutation, **kwargs)


class TTestRel(TTest):
    """Related measures t-test

    Notes
    -----
    For a t-test between two epochs, use an :class:`EpochCollection` epoch
    and ``model='epoch'``.
    """
    kind = 'ttest_rel'

    def __init__(self, model, c1, c0, tail=0):
        TTest.__init__(self, model, c1, c0, tail)
        assert self._between is None

    def make(self, y, ds, force_permutation, kwargs):
        return testnd.ttest_rel(
            y, self.model, self.c1, self.c0, 'subject', ds=ds, tail=self.tail,
            force_permutation=force_permutation, **kwargs)


class TContrastRel(EvokedTest):
    "T-contrast"
    kind = 't_contrast_rel'
    DICT_ATTRS = Test.DICT_ATTRS + ('contrast', 'tail')

    def __init__(self, model, contrast, tail=0):
        tail = tail_arg(tail)
        EvokedTest.__init__(contrast, model)
        self.contrast = contrast
        self.tail = tail

    def make(self, y, ds, force_permutation, kwargs):
        return testnd.t_contrast_rel(
            y, self.model, self.contrast, 'subject', ds=ds, tail=self.tail,
            force_permutation=force_permutation, **kwargs)


class ANOVA(EvokedTest):
    """ANOVA test

    Parameters
    ----------
    x : str
        ANOVA model specification (see :func:`test.anova`).
    model : str
        Model for grouping trials before averaging (does not need to be
        specified unless it should include variables not in ``x``).
    vars : tuple | dict
        Variables to add dynamically.
    """
    kind = 'anova'
    DICT_ATTRS = Test.DICT_ATTRS + ('x',)

    def __init__(self, x, model=None, vars=None):
        x = ''.join(x.split())
        if model is None:
            items = sorted(i.strip() for i in x.split('*'))
            within_items = (i for i in items if not re.match('^subject(\(\w+\))$', i))
            model = '%'.join(within_items)
        EvokedTest.__init__(self, x, model, vars=vars)
        if self._between is not None:
            raise NotImplementedError("Between-subject ANOVA")
        self.x = x

    def make(self, y, ds, force_permutation, kwargs):
        return testnd.anova(y, self.x, ds=ds, force_permutation=force_permutation, **kwargs)


class TwoStageTest(Test):
    "Two-stage test on epoched or evoked data"
    kind = 'two-stage'
    DICT_ATTRS = Test.DICT_ATTRS + ('stage_1',)

    def __init__(self, stage_1, vars=None, model=None):
        Test.__init__(self, stage_1, model, vars)
        self.stage_1 = stage_1

    def make_stage_1(self, y, ds, subject):
        """Assumes that model has already been applied"""
        return testnd.LM(y, self.stage_1, ds, subject=subject)

    def make_stage_2(self, lms, kwargs):
        lm = testnd.LMGroup(lms)
        lm.compute_column_ttests(**kwargs)
        return lm


TEST_CLASSES = {
    'anova': ANOVA,
    'ttest_1samp': TTestOneSample,
    'ttest_rel': TTestRel,
    'ttest_ind': TTestInd,
    't_contrast_rel': TContrastRel,
    'two-stage': TwoStageTest,
}
AGGREGATE_FUNCTIONS = ('mean', 'rms')
DATA_RE = re.compile("(source|sensor)(?:\.(%s))?$" % '|'.join(AGGREGATE_FUNCTIONS))


class TestDims(object):
    """Data shape for test

    Paremeters
    ----------
    string : str
        String describing data.
    time : bool
        Whether the base data contains a time axis.
    """
    source = None
    sensor = None

    def __init__(self, string, time=True):
        self.time = time
        substrings = string.split()
        for substring in substrings:
            m = DATA_RE.match(substring)
            if m is None:
                raise ValueError("Invalid test dimension description: %r" %
                                 (string,))
            dim, aggregate = m.groups()
            setattr(self, dim, aggregate or True)
        if sum(map(bool, (self.source, self.sensor))) != 1:
            raise ValueError("Invalid test dimension description: %r. Need "
                             "exactly one of 'sensor' or 'source'" % (string,))
        self.string = ' '.join(substrings)

        dims = []
        if self.source is True:
            dims.append('source')
        elif self.sensor is True:
            dims.append('sensor')
        if self.time is True:
            dims.append('time')
        self.dims = tuple(dims)

        # whether parc is used from subjects or from common-brain
        if self.source is True:
            self.parc_level = 'common'
        elif self.source:
            self.parc_level = 'individual'
        else:
            self.parc_level = None

    @classmethod
    def coerce(cls, obj, time=True):
        if isinstance(obj, cls):
            if bool(obj.time) == time:
                return obj
            else:
                return cls(obj.string, time)
        else:
            return cls(obj, time)

    def __repr__(self):
        return "TestDims(%r)" % (self.string,)

    def __eq__(self, other):
        if not isinstance(other, TestDims):
            return False
        return self.string == other.string and self.time == other.time


class ROITestResult(object):
    """Test results for temporal tests in one or more ROIs

    Attributes
    ----------
    subjects : tuple of str
        Subjects included in the test.
    samples : int
        ``samples`` parameter used for permutation tests.
    res : {str: NDTest} dict
        Test result for each ROI.
    n_trials_ds : Dataset
        Dataset describing how many trials were used in each condition per
        subject.
    """

    def __init__(self, subjects, samples, n_trials_ds, merged_dist, res):
        self.subjects = subjects
        self.samples = samples
        self.n_trials_ds = n_trials_ds
        self.merged_dist = merged_dist
        self.res = res

    def __getstate__(self):
        return {attr: getattr(self, attr) for attr in
                getargspec(self.__init__).args[1:]}

    def __setstate__(self, state):
        self.__init__(**state)



# Author: Christian Brodbeck <christianbrodbeck@nyu.edu>
#cython: boundscheck=False, wraparound=False

cimport cython
from cython.view cimport array as cvarray
from libc.stdlib cimport malloc, free
from libc.math cimport fabs
import numpy as np
cimport numpy as np

ctypedef np.int8_t INT8
ctypedef np.int64_t INT64
ctypedef np.float64_t FLOAT64


def l1(
        FLOAT64 [:] x,
        INT64 [:,:] indexes,
    ):
    cdef:
        double out = 0.
        size_t i, seg_i

    with nogil:
        for seg_i in range(indexes.shape[0]):
            for i in range(indexes[seg_i, 0], indexes[seg_i, 1]):
                out += fabs(x[i])

    return out


def l2(
        FLOAT64 [:] x,
        INT64 [:,:] indexes,
    ):
    cdef:
        double out = 0.
        size_t i, seg_i

    with nogil:
        for seg_i in range(indexes.shape[0]):
            for i in range(indexes[seg_i, 0], indexes[seg_i, 1]):
                out += x[i] ** 2

    return out


cdef void l1_for_delta(
        FLOAT64 [:] y_error,
        FLOAT64 [:] x,
        INT64 [:,:] indexes,  # training segment indexes
        double delta,
        size_t shift,
        double* e_add,
        double* e_sub,
    ) nogil:
    cdef:
        double d, temp_sum
        size_t i, seg_i, seg_start

    e_add[0] = 0.
    e_sub[0] = 0.

    for seg_i in range(indexes.shape[0]):
        seg_start = indexes[seg_i, 0]
        # start of the segment before the shift-delay
        temp_sum = 0.
        for i in range(seg_start, seg_start + shift):
            temp_sum += fabs(y_error[i])
        e_add[0] += temp_sum
        e_sub[0] += temp_sum
        for i in range(seg_start + shift, indexes[seg_i, 1]):
            d = delta * x[i - shift]
            e_add[0] += fabs(y_error[i] - d)
            e_sub[0] += fabs(y_error[i] + d)


cdef void l2_for_delta(
        FLOAT64 [:] y_error,
        FLOAT64 [:] x,
        INT64 [:,:] indexes,  # training segment indexes
        double delta,
        size_t shift,
        double* e_add,
        double* e_sub,
    ) nogil:
    cdef:
        double d, temp_sum
        size_t i, seg_i, seg_start

    e_add[0] = 0.
    e_sub[0] = 0.

    for seg_i in range(indexes.shape[0]):
        seg_start = indexes[seg_i, 0]
        # start of the segment before the shift-delay
        temp_sum = 0.
        for i in range(seg_start, seg_start + shift):
            temp_sum += y_error[i] ** 2
        e_add[0] += temp_sum
        e_sub[0] += temp_sum
        for i in range(seg_start + shift, indexes[seg_i, 1]):
            d = delta * x[i - shift]
            e_add[0] += (y_error[i] - d) ** 2
            e_sub[0] += (y_error[i] + d) ** 2


def generate_options(
        FLOAT64 [:] y_error,
        FLOAT64 [:,:] x,  # (n_stims, n_times)
        INT64 [:,:] indexes,  # training segment indexes
        int error,
        double delta,
        # buffers
        FLOAT64 [:,:] new_error,  # (n_stims, n_times_trf)
        INT8 [:,:] new_sign,
    ):
    cdef:
        double e_add
        double e_sub
        size_t n_stims = new_error.shape[0]
        size_t n_times_trf = new_error.shape[1]
        size_t i_stim, i_time
        FLOAT64 [:] x_stim

    if error != 1 and error != 2:
        raise RuntimeError("error=%r" % (error,))

    with nogil:
        for i_stim in range(n_stims):
            x_stim = x[i_stim]
            for i_time in range(n_times_trf):
                # +/- delta
                if error == 1:
                    l1_for_delta(y_error, x_stim, indexes, delta, i_time, &e_add, &e_sub)
                else:
                    l2_for_delta(y_error, x_stim, indexes, delta, i_time, &e_add, &e_sub)

                if e_add > e_sub:
                    new_error[i_stim, i_time] = e_sub
                    new_sign[i_stim, i_time] = -1
                else:
                    new_error[i_stim, i_time] = e_add
                    new_sign[i_stim, i_time] = 1


def update_error(
        FLOAT64 [:] y_error,
        FLOAT64 [:] x,
        INT64 [:,:] indexes,  # training segment indexes
        double delta,
        size_t shift,
    ):
    cdef:
        size_t i, seg_i

    with nogil:
        for seg_i in range(indexes.shape[0]):
            for i in range(indexes[seg_i, 0] + shift, indexes[seg_i, 1]):
                y_error[i] -= delta * x[i - shift]



# Author: Christian Brodbeck <christianbrodbeck@nyu.edu>
from contextlib import ContextDecorator
import os
import sys

IS_OSX = sys.platform == 'darwin'
IS_WINDOWS = os.name == 'nt'

if IS_OSX:
    from .macos import begin_activity, end_activity
else:
    from .dummy_os import begin_activity, end_activity


class Caffeinator(ContextDecorator):
    """Context disabling idle sleep and App Nap"""
    def __init__(self):
        self.n_processes = 0

    def __enter__(self):
        if self.n_processes == 0 and IS_OSX:
            self._activity = begin_activity()
        self.n_processes += 1

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.n_processes -= 1
        if self.n_processes == 0 and IS_OSX:
            end_activity(self._activity)


caffeine = Caffeinator()



"""Secondary variables"""
from .._data_obj import Var, asfactor

import numpy as np

from .definitions import DefinitionError


def as_vardef_var(v):
    "Coerce ds.eval() output for use as variable"
    if isinstance(v, np.ndarray):
        if v.dtype.kind == 'b':
            return Var(v.astype(int))
        return Var(v)
    return v


class VarDef(object):

    def apply(self, ds, e):
        raise NotImplementedError


class EvalVar(VarDef):

    def __init__(self, code):
        assert isinstance(code, str)
        self.code = code

    def __repr__(self):
        return "EvalVar(%r)" % self.code

    def __eq__(self, other):
        return isinstance(other, EvalVar) and other.code == self.code

    def apply(self, ds, e):
        return as_vardef_var(ds.eval(self.code))


class LabelVar(VarDef):

    def __init__(self, source, codes):
        self.source = source
        self.codes = codes

    def __repr__(self):
        return "LabelVar(%r, %r)" % (self.source, self.codes)

    def __eq__(self, other):
        return (isinstance(other, LabelVar) and other.source == self.source
                and other.codes == self.codes)

    def apply(self, ds, e):
        v = asfactor(self.source, ds=ds).as_var(self.codes, 0)
        return as_vardef_var(v)


class GroupVar(VarDef):
    """Group membership for each subject

    Parameters
    ----------
    groups : tuple | dict
        Groups to consider. A tuple of group names to lookup for each subject
        which of those groups it belongs to. A {group: label} dict to assign
        a label based on group membership.
    """

    def __init__(self, groups):
        self.groups = groups

    def __repr__(self):
        return "GroupVar(%r)" % (self.groups,)

    def __eq__(self, other):
        return isinstance(other, GroupVar) and other.groups == self.groups

    def apply(self, ds, e):
        return e.label_groups(ds['subject'], self.groups)

    @classmethod
    def from_string(cls, string):
        groups = {}
        for item in string.split(','):
            if ':' in item:
                src, dst = map(str.strip, item.split(':'))
            else:
                src = dst = item.strip()
            groups[src] = dst
        if all(k == v for k, v in groups.items()):
            groups = tuple(sorted(groups))
        return cls(groups)


def parse_named_vardef(string):
    if '=' not in string:
        raise DefinitionError("var %r: needs '='" % (string,))
    name, vdef = string.split('=', 1)
    return name.strip(), parse_vardef(vdef)


def parse_vardef(string):
    string = string.strip()
    if string.startswith('group:'):
        return GroupVar.from_string(string[6:])
    else:
        return EvalVar(string)


class Vars(object):
    """Set of variable definitions

    Parameters
    ----------
    arg : str | tuple | dict
        The ``vars`` argument.
    """

    def __init__(self, arg):
        if isinstance(arg, str):
            arg = (arg,)
        elif isinstance(arg, dict):
            arg = arg.items()
        elif not isinstance(arg, (tuple, list)):
            raise TypeError("vars=%s; needs to be atrs, dict or tuple" % (arg,))

        items = []
        for item in arg:
            if isinstance(item, str):
                name, vdef = parse_named_vardef(item)
            else:
                name, vdef = item
                if isinstance(vdef, str):
                    vdef = parse_vardef(vdef)
                elif isinstance(vdef, VarDef):
                    pass
                else:
                    vdef = LabelVar(*vdef)
            items.append((name, vdef))
        self.items = items

    def __repr__(self):
        return "Vars(%s)" % (self.items,)

    def __eq__(self, other):
        return isinstance(other, Vars) and other.items == self.items

    def apply(self, ds, e):
        for name, vdef in self.items:
            ds[name] = vdef.apply(ds, e)



import re

from .definitions import DefinitionError, Definition


COMBINATION_PARC = 'combination'
FS_PARC = 'subject_parc'  # Parcellation that come with every MRI-subject
FSA_PARC = 'fsaverage_parc'  # Parcellation that comes with fsaverage
SEEDED_PARC = 'seeded'
INDIVIDUAL_SEEDED_PARC = 'individual seeded'
SEEDED_PARC_RE = re.compile('^(.+)-(\d+)$')


class Parcellation(Definition):
    DICT_ATTRS = ('kind',)
    make = False
    morph_from_fsaverage = False

    def __init__(self, name, views=None):
        self.name = name
        self.views = views


class CombinationParcellation(Parcellation):
    "Recombine labels from an existing parcellation"
    DICT_ATTRS = ('kind', 'base', 'labels')
    kind = COMBINATION_PARC
    make = True

    def __init__(self, name, base, labels, views=None):
        Parcellation.__init__(self, name, views)
        self.base = base
        self.labels = labels


class EelbrainParcellation(Parcellation):
    "Parcellation that has special make rule"
    kind = 'eelbrain_parc'
    make = True

    def __init__(self, name, morph_from_fsaverage, views=None):
        Parcellation.__init__(self, name, views)
        self.morph_from_fsaverage = morph_from_fsaverage


class FreeSurferParcellation(Parcellation):
    "Parcellation that comes with FreeSurfer"
    kind = FS_PARC


class FSAverageParcellation(Parcellation):
    "Parcellation that comes with FSAverage"
    kind = FSA_PARC
    morph_from_fsaverage = True


class LabelParcellation(Parcellation):
    "Assemble parcellation from labels"
    DICT_ATTRS = ('kind', 'labels')
    kind = 'label_parc'
    make = True

    def __init__(self, name, labels, views=None):
        Parcellation.__init__(self, name, views)
        self.labels = labels if isinstance(labels, tuple) else tuple(labels)


class SeededParcellation(Parcellation):
    "Parcellation that is grown from seed vertices"
    DICT_ATTRS = ('kind', 'seeds', 'surface', 'mask')
    kind = SEEDED_PARC
    make = True

    def __init__(self, name, seeds, mask=None, surface='white', views=None):
        Parcellation.__init__(self, name, views)
        self.seeds = seeds
        self.mask = mask
        self.surface = surface

    def seeds_for_subject(self, subject):
        return self.seeds


class IndividualSeededParcellation(SeededParcellation):
    "Seed parcellation with individual seeds for each subject"
    kind = INDIVIDUAL_SEEDED_PARC
    morph_from_fsaverage = False

    def __init__(self, name, seeds, mask=None, surface='white', views=None):
        SeededParcellation.__init__(self, name, seeds, mask, surface, views)
        labels = tuple(self.seeds)
        label_subjects = {label: sorted(self.seeds[label].keys()) for label in labels}
        subjects = label_subjects[labels[0]]
        if not all(label_subjects[label] == subjects for label in labels[1:]):
            raise DefinitionError(
                "parc %s: Some labels are missing subjects. if a subject does not "
                "have a label, use an empty tuple as seed, e.g.: 'R0001': ()" % self.name)
        self.subjects = subjects

    def seeds_for_subject(self, subject):
        if subject not in self.subjects:
            raise DefinitionError("Parc %s not defined for subject %s" % (self.name, subject))
        seeds = {name: self.seeds[name][subject] for name in self.seeds}
        # filter out missing
        return {name: seed for name, seed in seeds.items() if seed}


PARC_CLASSES = {
    COMBINATION_PARC: CombinationParcellation,
    FS_PARC: FreeSurferParcellation,
    FSA_PARC: FSAverageParcellation,
    SEEDED_PARC: SeededParcellation,
    INDIVIDUAL_SEEDED_PARC: IndividualSeededParcellation,
}
